class alttprException(Exception):
    pass


class alttprFailedToRetrieve(Exception):
    pass


class alttprFailedToGenerate(Exception):
    pass


class UnableToRetrieve(Exception):
    pass

class UnableToGenerate(Exception):
    pass